import System.CPUTime
import Control.Monad
import Data.List ((\\), replicate)
import Text.Printf
import Test.QuickCheck
 
-- Sorting


-- isort: insertion sort

isort :: Ord a => [a] -> [a]
isort = foldr insert []
  where
  insert :: Ord a => a -> [a] -> [a]
  insert x []                    =  [x]
  insert x (y : ys) | x <= y     =  x : y : ys
                    | otherwise  =  y : insert x ys

-- qsort: quicksort

qsort :: Ord a => Int -> [a] -> [a]
qsort k xs | length xs <= k  =  isort xs
qsort k (y:xs) =
  qsort k [ x | x <- xs, x < y ]
  ++ [ y ] ++
  qsort k [ x | x <- xs, x >= y ]

-- msort: merge sort

msort :: Ord a => Int -> [a] -> [a]
msort k xs | length xs <= k  =  isort xs
           | otherwise       =  merge (msort k (take m xs))
                                      (msort k (drop m xs))
  where
  m = length xs `div` 2
  merge :: Ord a => [a] -> [a] -> [a]
  merge xs []                      =  xs
  merge [] ys                      =  ys
  merge (x:xs) (y:ys) | x <= y     =  x : merge xs (y:ys)
                      | otherwise  =  y : merge (x:xs) ys

-- Lexical order

(<@) :: Ord a => [a] -> [a] -> Bool
xs     <@ []               =  False
[]     <@ (y:ys)           =  True
(x:xs) <@ (y:ys) | x < y   =  True
                 | x == y  =  xs <@ ys
                 | x > y   =  False

(@>) :: Ord a => [a] -> [a] -> Bool
xs @> ys  =  ys <@ xs

prop_trichotomy :: String -> String -> Bool
prop_trichotomy xs ys  =
  exactlyOne [ xs <@ ys , xs == ys , xs @> ys ]
  where
  exactlyOne :: [Bool] -> Bool
  exactlyOne xs  =  length (filter id xs) == 1

-- Timing

type Time = Double

-- time
-- returns time to compute a function applied to its argument, in seconds

time :: ([a] -> [b]) -> [a] -> IO Time
time f xs =
  do
    xs `seq` return ()
    start <- getCPUTime
    length (f xs) `seq` return ()
    end   <- getCPUTime
    return (fromIntegral (end - start) / 10^12)

-- timer
-- returns time to compute a function applied to a random
-- list of length m

random :: Arbitrary a => Int -> IO [a]
random m  =  generate (sequence (replicate m arbitrary))

timera :: Arbitrary a => ([a] -> [b]) -> Int -> IO Time
timera f m =
  do
    xs <- random m
    time f xs

timer :: ([Int] -> [Int]) -> Int -> IO Time
timer =  timera

-- run
-- returns average of n times, with minimum and maximum times removed

run :: Int -> IO Time -> IO Time
run n act =
  do
    ts <- sequence (replicate n act)
    return (average (inliers ts))

average :: Fractional a => [a] -> a
average ts  =  sum ts / fromIntegral (length ts)

inliers :: Ord a => [a] -> [a]
inliers ts  =  ts \\ [maximum ts, minimum ts]

-- Testing data

countdown :: Int -> [Int]
countdown n  =  [n, n-1 .. 1]

psequence :: [IO Time] -> IO ()
psequence mts =
  do
    ts <- sequence mts
    sequence [ printf "%.6f\n" t | t <- ts ]
    return ()

-- time as a function of size of input, random

randi, randq, randm :: IO ()
randi =
  psequence
    [ run 10 (timer isort m)
    | m <- [0,500..5000] ]
randq =
  psequence
    [ run 10 (timer (qsort 30) m)
    | m <- [0,500..5000] ]
randm =
  psequence
    [ run 10 (timer (msort 30) m)
    | m <- [0,500..5000] ]


-- time as a function of size of input, reverse ordered

downi, downq, downm :: IO ()
downi =
  psequence
    [ run 10 (time isort (countdown m))
    | m <- [0,500..5000] ]
downq =
  psequence
    [ run 10 (time (qsort 30) (countdown m))
    | m <- [0,500..5000] ]
downm =
  psequence
    [ run 100 (time (msort 30) (countdown m))
    | m <- [0,500..5000] ]

-- time as a function of cutoff

cutq, cutm :: IO ()
cutq =
  psequence
    [ run 100 (timer (qsort i) (2^10))
    | i <- [0,10..100] ]
cutm =
  psequence
    [ run 100 (timer (msort i) (2^10))
    | i <- [1,11..101] ]
